!-------------------------------------------------------------------------------
! CDSDv v.1.2
! Copyright 2020 João Vargas
!
! This program is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser Public License for more details.
!
! You should have received a copy of the GNU Lesser Public License
! along with this program.  If not, see <http://www.gnu.org/licenses/>.
!
! Last Update: João Vargas   23/07/2020
!-------------------------------------------------------------------------------
! This work should be cited as J. Vargas, B. Lopez, M. Lino da Silva, 
! "CDSDv: A Compact Database for the Modelling of High-Temperature CO2 Radiation"
! Journal of Quantitative Spectroscopy and Radiative Transfer, accepted for publication
!-------------------------------------------------------------------------------
Program Main


  implicit none
  integer       ,parameter        :: rp       = 8               ! Real Precision
  real(rp)      ,parameter        :: c2       = 1.438776877E-2  ! second radiation constant K/m-1
  real(rp)      ,parameter        :: c0       = 299792458.0_rp  ! speed of light, m/s
  real(rp)      ,parameter        :: kB       = 1.380649E-23    ! Boltzmann constant J/K
  real(rp)      ,parameter        :: hP       = 6.62607015E-34  ! Planck constant J/s
  real(rp)      ,parameter        :: Ru       = 8.314462618_rp  ! Universal gas constant, J/K/mol
  real(rp)      ,parameter        :: Na       = 6.0221409E+23   ! Avogadro's number
  real(rp)      ,parameter        :: pi       = 3.1415927_rp    ! Pi
  !# Inputs
  character(*)  ,parameter        :: InFile   = 'input.txt'
  character(:)  ,allocatable      :: LevFile                    ! Levels file
  character(:)  ,allocatable      :: TraFile                    ! Transition file
  integer                         :: Jmax                       ! Maximum rotational number
  real(rp)                        :: Tg                         ! Translational/gas temperature, Kelvin
  real(rp)                        :: Te                         ! Excitation temperature, Kelvin
  real(rp)                        :: Tr                         ! Rotational temperature, Kelvin
  real(rp)                        :: Tv                         ! Vibrational temperature, Kelvin
  real(rp)                        :: P                          ! Gas pressure, Pascal
  integer                         :: Nspe                       ! Number of species in gas
  real(rp)      ,allocatable      :: X(:)                       ! Mole fractions, size(X) = Nspe
  real(rp)      ,allocatable      :: M(:)                       ! Molar masses, size(M) = Nspe, Kg/mol
  real(rp)      ,allocatable      :: R(:)                       ! Molecular/Atomic radius, size(R) = Nspe, Angstrom
  real(rp)                        :: LW                         ! Lower limit of spectroscopic window, cm-1
  real(rp)                        :: UW                         ! Upper limit of spectroscopic window, cm-1
  logical                         :: NonEqInd                   ! Indicator of non-equilibrium input
  character(:)  ,allocatable      :: NonEqFile                  ! File with non-equilibrium mole fractions
  real(rp)                        :: DecimParam                 ! At the end, this decimation parameter will not output lines
                                                                ! which are below the maximum values of emission and absorption
                                                                ! like so: if( line/maxval < decimparam ) forget; else output
                                                                
  !# Intermediary variables       
  real(rp)                        :: NoV                        ! Number of particles per unit of volume, m-3
  integer                         :: NLev                       ! Number of levels in database
  integer                         :: NTra
  integer       ,allocatable      :: VibAssign(:,:)             ! Vibrational assignment of levels, size(VibAssign) = (NLev,5), each vibrational level is made explicit 
                                                                ! with 5 vibrational numbers as follows: (v1,v2,l2,v3,r)
                                                                ! v1 - symmetric stretch
                                                                ! v2 - bending
                                                                ! l2 - vibrational angular momentum
                                                                ! v3 - asymmetric stretch
                                                                ! r  - ranking number
  real(rp)      ,allocatable      :: Ecoe(:,:)                  ! Coefficients for polynomial expansion of energy levels, size(Ecoe) = (NLev,4)
                                                                ! Ev = Ecoe(v,1) + Ecoe(v,2)*J(J+1) - Ecoe(v,3)*(J(J+1))^2 + Ecoe(v,4)*(J(J+1))^3
  integer       ,allocatable      :: LevJ(:,:)                  ! Limits of rotational numbers for polynomial expansions, size(LevJ) = (NLev,2)
  real(rp)      ,allocatable      :: LevPop(:)                  ! Number density of each level organized by the same order it is read in the LevFile
  real(rp)                        :: Qtot                       ! Ro-vibrational partition function of CO2, can be approximated to Qtot in respect to IR radiation
  integer       ,allocatable      :: TraAssign(:,:)             ! Vibrational transition assignment, size(TraAssign) = (NTra,10)
                                                                ! Fully determined between 2 vibrational level assignments
  real(rp)      ,allocatable      :: VibTraCoe(:,:)             ! Vibrational transition coefficients, size(VibTraCoe) = (NTra,5)
                                                                ! VibTraCoe(v,1) is the vibrational Einstein coefficient Avv in s-1
                                                                ! VibTraCoe(v,2:5) are the Herman-Wallis coefficients respectively, A1, A2, A3, AQ
  integer       ,allocatable      :: LevTraIndex(:,:)           ! Indexes of upper and lower vibrational levels, size(LevTraIndex) = (NTra,2)
  integer                         :: NTotTra                    ! Total number of ro-vibrational transitions 
  logical       ,allocatable      :: ProcRoVibTrans(:)          ! Indicator if vibrational transition contains rovib transitions inside the spectral region in calculation
  real(rp)      ,allocatable      :: LineParam(:,:)             ! Line parameters for output calculation size(LineParam) = (NTotTra,6)
                                                                ! 1  - Line Position, cm-1
                                                                ! 2  - Ro-vibrational Einstein coefficient, s-1
                                                                ! 3  - Upper level degeneracy
                                                                ! 4  - Upper level ro-vibrational partition function contribution divided by the degeneracy
                                                                ! 5  - Lower level degeneracy
                                                                ! 6  - Lower level ro-vibrational partition function contribution divided by the degeneracy                                                                ! In the case of Non-equilibrium input, meaning the specification of vibrational specific number densities 
                                                                ! the 4th and 6h parameter change accordingly:
                                                                ! 4  - Vibrational number density of the upper level x upper rotational level fraction x total partition function
                                                                !      divided by the upper rotational degeneracy and the sum of all vibrational number densities
                                                                ! 6  - Vibrational number density of the lower level x lower rotational level fraction x total partition function
                                                                !      divided by the lower rotational degeneracy and the sum of all vibrational number densities
                                                                ! The definition of the 4 and 6 line parameters in the non-equilibrium regime does not require modifications
                                                                ! of the code further on, remaining valid for non-equilibrium and equilibrium modes of the code
                                                                
  !# Outputs, all the same size = NTotTra                                 
  real(rp)      ,allocatable      :: nu(:)                      ! Line position, cm-1
  real(rp)      ,allocatable      :: Ie(:)                      ! Emission coefficient, W/m3/sr/cm-1
  real(rp)      ,allocatable      :: Ia(:)                      ! Absorption coefficient, m-1
  real(rp)      ,allocatable      :: DnuL(:)                    ! FWHM Lorentzian/collision broadening, cm-1
  real(rp)      ,allocatable      :: DnuD(:)                    ! FWHM Doppler broadening, cm-1

  write(*,*) 'CDSDvReader greets you!'
  write(*,*) 'Reading input data'
  call ReadInput( InFile, LevFile, TraFile, Jmax, Tg, Te, Tr, Tv, P, Nspe, X, M, R, LW, UW, NonEqInd, NonEqFile, DecimParam )
  NoV = P/(kB*Tg)
  
  write(*,*) 'Input data has been read, reading CDSDv'
  call ReadLevelData( LevFile, Tv, Tr, NLev, Qtot, VibAssign, Ecoe, LevJ )
  call ReadTransData( TraFile, NTra, TraAssign, VibTraCoe )
  
  if( NonEqInd ) then
    call ReadNonEquilibrium( NLev, NonEqFile, LevPop )
    NoV = sum(LevPop)
  end if
  
  write(*,*) 'CDSDv has been read, developing ro-vibrational data'
  call LinkLevelsTransitions( NLev, NTra, VibAssign, TraAssign, LevTraIndex )
  
  write(*,*) 'Levels and transitions have been linked, counting total number of transitions'
  call CountRoVibTrans( Jmax, LW, UW, NLev, NTra, VibAssign, LevTraIndex, Ecoe, LevJ, NTotTra, ProcRoVibTrans )
  
  write(*,*) 'Transitions have been counted, computing parameters'
  call InitializeRoVibTrans( Jmax, LW, UW, NLev, NTra, VibAssign, LevTraIndex, Ecoe, LevJ, VibTraCoe, &
                             NTotTra, ProcRoVibTrans, Tr, Tv, Qtot, NonEqInd, LevPop, LineParam )

  write(*,*) 'Computing output'
  call ComputeRoVibTrans( NTotTra, LineParam, Qtot, Tg, NoV, Nspe, X, M, R, nu, Ie, Ia, DnuL, DnuD )
  
  write(*,*) 'Decimating and Outputting'
  call OutputData( NTotTra, nu, Ie, Ia, DnuL, DnuD, DecimParam )
  
  contains
  
Subroutine ReadInput( InFile, LevFile, TraFile, Jmax, Tg, Te, Tr, Tv, P, Nspe, X, M, R, LW, UW, NonEqInd, NonEqFile, DecimParam )
  character(*)                    ,intent(in)       :: InFile
  character(:)  ,allocatable      ,intent(out)      :: LevFile
  character(:)  ,allocatable      ,intent(out)      :: TraFile
  integer                         ,intent(out)      :: Jmax
  real(rp)                        ,intent(out)      :: Tg
  real(rp)                        ,intent(out)      :: Te
  real(rp)                        ,intent(out)      :: Tr
  real(rp)                        ,intent(out)      :: Tv
  real(rp)                        ,intent(out)      :: P
  integer                         ,intent(out)      :: Nspe
  real(rp)      ,allocatable      ,intent(out)      :: X(:)
  real(rp)      ,allocatable      ,intent(out)      :: M(:)
  real(rp)      ,allocatable      ,intent(out)      :: R(:)
  real(rp)                        ,intent(out)      :: LW
  real(rp)                        ,intent(out)      :: UW
  logical                         ,intent(out)      :: NonEqInd
  character(:)  ,allocatable      ,intent(out)      :: NonEqFile
  real(rp)                        ,intent(out)      :: DecimParam
  integer                                           :: Unit
  character(100)                                    :: String
  character(*)                    ,parameter        :: Separator = " "
  integer                                           :: iSep
  
  open( Newunit=Unit, File=InFile, Status='OLD', Action='READ' )
  
  read(Unit,'(a)') String
  iSep      = index(String,Separator)
  LevFile   = trim(adjustl(String(1:iSep)))
  read(Unit,'(a)') String
  iSep      = index(String,Separator)
  TraFile   = trim(adjustl(String(1:iSep)))
  
  read(Unit,'(i5)') Jmax
  read(Unit,'(es15.8)') Tg
  read(Unit,'(es15.8)') Te
  read(Unit,'(es15.8)') Tr
  read(Unit,'(es15.8)') Tv
  read(Unit,'(es15.8)') P
  read(Unit,'(i5)') Nspe
  
  allocate(X(Nspe))
  allocate(M(Nspe))
  allocate(R(Nspe))
  
  read(Unit,*) X  
  read(Unit,*) M
  read(Unit,*) R
  read(Unit,'(es15.8)') LW
  read(Unit,'(es15.8)') UW
  read(Unit,'(L)') NonEqInd
  read(Unit,'(a)') String
  iSep      = index(String,Separator)
  NonEqFile = trim(adjustl(String(1:iSep)))
  read(Unit,'(es15.8)') DecimParam
  
  
  close(Unit)

End Subroutine

Subroutine ReadLevelData( InFile, Tv, Tr, NLev, Qtot, VibAssign, Ecoe, LevJ )
  character(*)                    ,intent(in)       :: InFile
  real(rp)                        ,intent(in)       :: Tv
  real(rp)                        ,intent(in)       :: Tr
  integer                         ,intent(out)      :: NLev
  real(rp)                        ,intent(out)      :: Qtot
  integer       ,allocatable      ,intent(out)      :: VibAssign(:,:)
  real(rp)      ,allocatable      ,intent(out)      :: ECoe(:,:)
  integer       ,allocatable      ,intent(out)      :: LevJ(:,:)
  real(rp)      ,parameter                          :: c2_cmK = 100.0_rp*c2
  integer                                           :: Unit
  integer                                           :: iLev
  real(rp)                                          :: Qv
  real(rp)                                          :: gv
  real(rp)                                          :: Qr
  real(rp)                                          :: The
  real(rp)                                          :: The1
  
  open( Newunit=Unit, File=InFile, Status='OLD', Action='READ')
  read(Unit,'(i5)') NLev
  read(Unit,*) ! Skip header line

  allocate(VibAssign(NLev,5))
  allocate(ECoe(NLev,4))
  allocate(LevJ(NLev,2))
  Qtot    = 0.0_rp
  do iLev = 1,NLev
    read(Unit,'(5i3,4es18.10,2i4)') VibAssign(iLev,:), ECoe(iLev,:), LevJ(iLev,:)
    if( VibAssign(iLev,3) > 0 ) then
      gv  = 2.0_rp
    else
      gv  = 1.0_rp
    end if
    ! Here the vibrational angular momentum must be considered as the approximation used for the rotational
    ! partition function does not consider rotational degeneracy, as such gv = 1 if l2 = 0 and gv = 2 if l2 > 0
    Qv    = gv*exp(-c2_cmK*ECoe(iLev,1)/Tv)
    ! Approximation for Qr ~ Tr/(sigma*theta)*exp( theta/(3*Tr) + theta^2/(90*Tr^2) ) when theta << Tr. 
    ! theta = c2_cmK*B_{v=0}, sigma = 2 for symmetric molecules
    ! It is assumed that the first level in database is (0,0,0,0,0), hence The = Tr/(c2_cmK*Ecoe(1,2))
    The   = Tr/(c2_cmK*ECoe(1,2))
    The1  = 1.0/The
    Qr    = (The/2.0_rp)*exp( The1/3.0_rp + The1*The1/90.0 )
    Qtot  = Qtot + Qv*Qr
  end do
  write(*,"(a35,es11.4)") 'Ro-vibrational partition function:',Qtot 
  close(Unit)
End Subroutine
  
Subroutine ReadTransData(InFile, NTra, TraAssign, VibTraCoe )
  character(*)                    ,intent(in)       :: InFile
  integer                         ,intent(out)      :: NTra
  integer       ,allocatable      ,intent(out)      :: TraAssign(:,:)
  real(rp)      ,allocatable      ,intent(out)      :: VibTraCoe(:,:)
  integer                                           :: Unit
  integer                                           :: iTra
  
  open( Newunit=Unit, File=InFile, Status='OLD', Action='READ' )
  read(Unit,'(i8)') NTra
  read(Unit,*) ! Skip header line
  allocate(TraAssign(NTra,10))
  allocate(VibTraCoe(NTra,5))
  do iTra = 1,NTra
    read(Unit,'(10i3,5es16.8)') TraAssign(iTra,:), VibTraCoe(iTra,:)
  end do
  close(Unit)
End Subroutine

Subroutine ReadNonEquilibrium( NLev, NonEqFile, LevPop )
  integer                         ,intent(in)       :: NLev
  character(*)                    ,intent(in)       :: NonEqFile
  real(rp)      ,allocatable      ,intent(out)      :: LevPop(:)
  integer                                           :: Unit
  integer                                           :: iLev
  
  allocate(LevPop(NLev))
  open( Newunit=Unit, File=NonEqFile, Status='OLD', Action='READ' )
  do iLev = 1,NLev
    read(Unit,'(es16.8e3)') LevPop(iLev)
  end do
  close(Unit)

End Subroutine
  
Subroutine LinkLevelsTransitions( NLev, NTra, VibAssign, TraAssign, LevTraIndex )
  integer                         ,intent(in)       :: NLev
  integer                         ,intent(in)       :: NTra
  integer                         ,intent(in)       :: VibAssign(NLev,5)
  integer                         ,intent(in)       :: TraAssign(NTra,10)
  integer       ,allocatable      ,intent(out)      :: LevTraIndex(:,:)
  integer                                           :: iLev
  integer                                           :: iTra
  integer                                           :: vNum
  
  allocate(LevTraIndex(NTra,2))
  LevTraIndex   = 0
  
  do iTra = 1,NTra
    do iLev = 1,NLev
    
      UpperLevel: Block
        if( LevTraIndex(iTra,1) == 0 ) then
          do vNum = 1,5
            if( VibAssign(iLev,vNum) /= TraAssign(iTra,vNum) ) exit UpperLevel
          end do
          LevTraIndex(iTra,1) = iLev
        end if
      End Block UpperLevel
      
      LowerLevel: Block
        if( LevTraIndex(iTra,2) == 0 ) then
          do vNum = 6,10
            if( VibAssign(iLev,vNum-5) /= TraAssign(iTra,vNum) ) exit LowerLevel
          end do
          LevTraIndex(iTra,2) = iLev
        end if
      End Block LowerLevel
    
    end do
  end do

End Subroutine

Subroutine CountRoVibTrans( Jmax, LW, UW, NLev, NTra, VibAssign, LevTraIndex, Ecoe, LevJ, NTotTra, ProcRoVibTrans )
  integer                         ,intent(in)       :: Jmax
  real(rp)                        ,intent(in)       :: LW
  real(rp)                        ,intent(in)       :: UW
  integer                         ,intent(in)       :: NLev
  integer                         ,intent(in)       :: NTra
  integer                         ,intent(in)       :: VibAssign(NLev,5)
  integer                         ,intent(in)       :: LevTraIndex(NTra,2)
  real(rp)                        ,intent(in)       :: Ecoe(NLev,4)
  integer                         ,intent(in)       :: LevJ(NLev,2)
  integer                         ,intent(out)      :: NTotTra
  logical       ,allocatable      ,intent(out)      :: ProcRoVibTrans(:)
  integer                                           :: iTra
  logical                                           :: Over
  logical                                           :: NotFound
  logical                                           :: Test
  integer                                           :: Vlow
  integer                                           :: Vupp
  integer                                           :: Jlow
  integer                                           :: Jupp
  integer                                           :: DeltaJ
  real(rp)                                          :: Eupp
  real(rp)                                          :: Elow
  real(rp)                                          :: LinePos
  integer       ,allocatable                        :: NRoVibTrans(:)
  
  allocate(NRoVibTrans(NTra))
  allocate(ProcRoVibTrans(NTra))
  ProcRoVibTrans(:)   = .false.
  
  do iTra = 1,NTra
    Over              = .false.
    NotFound          = .true.
    NRoVibTrans(iTra) = 0
    Vlow              = LevTraIndex(iTra,2)
    Vupp              = LevTraIndex(iTra,1)
    Jlow              = LevJ(Vlow,1)
    Jupp              = LevJ(Vupp,1)
    
    do while( .not. Over )
      if( Jupp > LevJ(Vupp,2) .OR. Jlow > LevJ(Vlow,2) .OR. Jupp > Jmax .OR. Jlow > Jmax ) then
        Over            = .true.
        cycle
      end if
      DeltaJ          = Jupp - Jlow
      Eupp            = Energy( Ecoe(Vupp,:), Jupp )
      Elow            = Energy( Ecoe(Vlow,:), Jlow )
      LinePos         = Eupp - Elow
      Test            = ( LW <= LinePos .AND. LinePos <= UW )
      select case(DeltaJ)
        case( 2: )
          Jlow                  = ChangeJ( Jlow, VibAssign(Vlow,3), 1 )
        case( 1, 0, -1 )
          ProcRoVibTrans(iTra)  = .true.
          if( Test ) then
            NRoVibTrans(iTra)   = NRoVibTrans(iTra) + 1 
            NotFound            = .false.
          end if
          Jlow                  = ChangeJ( Jlow, VibAssign(Vlow,3), 1 )
        case( :-2 )
          Jupp                  = ChangeJ( Jupp, VibAssign(Vupp,3), 1 )
          Jlow                  = ChangeJ( Jlow, VibAssign(Vlow,3), -2 )
          if( NotFound .OR. Jlow < 0 ) Jlow = LevJ(Vlow,1)
      end select
    end do
  end do
  
  NTotTra = sum(NRoVibTrans)
  
End Subroutine

Subroutine InitializeRoVibTrans( Jmax, LW, UW, NLev, NTra, VibAssign, LevTraIndex, Ecoe, LevJ, VibTraCoe, &
                                 NTotTra, ProcRoVibTrans, Tr, Tv, Qtot, NonEqInd, LevPop, LineParam )
  integer                         ,intent(in)       :: Jmax
  real(rp)                        ,intent(in)       :: LW
  real(rp)                        ,intent(in)       :: UW
  integer                         ,intent(in)       :: NLev
  integer                         ,intent(in)       :: NTra
  integer                         ,intent(in)       :: VibAssign(NLev,5)
  integer                         ,intent(in)       :: LevTraIndex(NTra,2)
  real(rp)                        ,intent(in)       :: Ecoe(NLev,4)
  integer                         ,intent(in)       :: LevJ(NLev,2)
  real(rp)                        ,intent(in)       :: VibTraCoe(NTra,5)
  integer                         ,intent(in)       :: NTotTra
  logical       ,allocatable      ,intent(in)       :: ProcRoVibTrans(:)
  real(rp)                        ,intent(in)       :: Tr
  real(rp)                        ,intent(in)       :: Tv
  real(rp)                        ,intent(in)       :: Qtot
  logical                         ,intent(in)       :: NonEqInd
  real(rp)    ,allocatable        ,intent(in)       :: LevPop(:)
  real(rp)    ,allocatable        ,intent(out)      :: LineParam(:,:)
  integer                                           :: iTra
  logical                                           :: Over
  logical                                           :: NotFound
  logical                                           :: Test
  integer                                           :: Vlow
  integer                                           :: Vupp
  integer                                           :: Jlow
  integer                                           :: Jupp
  integer                                           :: l2low
  integer                                           :: l2upp
  integer                                           :: DeltaJ
  integer                                           :: iTot
  real(rp)                                          :: Eupp
  real(rp)                                          :: Nup
  real(rp)                                          :: Elow
  real(rp)                                          :: Nlo
  real(rp)                                          :: LinePos
  real(rp)                                          :: VibParam(5)
  
  iTot = 1
  allocate(LineParam(NTotTra,6))
  
  do iTra = 1,NTra
    if( .not. ProcRoVibTrans(iTra) ) cycle
    Over              = .false.
    NotFound          = .true.
    Vlow              = LevTraIndex(iTra,2)
    Vupp              = LevTraIndex(iTra,1)
    Jlow              = LevJ(Vlow,1)
    Jupp              = LevJ(Vupp,1)
    l2low             = VibAssign(Vlow,3)
    l2upp             = VibAssign(Vupp,3)
    VibParam          = VibTraCoe(iTra,:)

    do while( .not. Over )
      if( Jupp > LevJ(Vupp,2) .OR. Jlow > LevJ(Vlow,2) .OR. Jupp > Jmax .OR. Jlow > Jmax ) then
        Over            = .true.
        cycle
      end if
      DeltaJ          = Jupp - Jlow
      Eupp            = Energy( Ecoe(Vupp,:), Jupp )
      Elow            = Energy( Ecoe(Vlow,:), Jlow )
      LinePos         = Eupp - Elow
      Test            = ( LW <= LinePos .AND. LinePos <= UW )
      select case(DeltaJ)
        case( 2: )
          Jlow                = ChangeJ( Jlow, l2low, 1 )
        case( 1, 0, -1 )
          if( Test ) then
            if( NonEqInd ) then
              Nup             = LevPop(Vupp)
              Nlo             = LevPop(Vlow)
            end if
            LineParam(iTot,:) = ComputeLineParam( LinePos, VibParam, Jupp, l2upp, Ecoe(Vupp,:), Jlow, l2low, &
                                                  Ecoe(Vlow,:), Tr, Tv, NonEqInd, Nup, Nlo, Qtot )
            iTot              = iTot + 1
            NotFound          = .false.
          end if
          Jlow                = ChangeJ( Jlow, l2low, 1 )
        case( :-2 )
          Jupp                = ChangeJ( Jupp, l2upp, 1 )
          Jlow                = ChangeJ( Jlow, l2low, -2 )
          if( NotFound .OR. Jlow < 0 ) Jlow = LevJ(Vlow,1)
      end select
    end do
  end do

End Subroutine

Subroutine ComputeRoVibTrans( NTotTra, LineParam, Qtot, Tg, NoV, Nspe, X, M, R, nu, Ie, Ia, DnuL, DnuD )
  integer                         ,intent(in)       :: NTotTra
  real(rp)                        ,intent(in)       :: LineParam(NTotTra,6)
  real(rp)                        ,intent(in)       :: Qtot
  real(rp)                        ,intent(in)       :: Tg
  real(rp)                        ,intent(in)       :: NoV
  integer                         ,intent(in)       :: Nspe
  real(rp)                        ,intent(in)       :: X(Nspe)
  real(rp)                        ,intent(in)       :: M(Nspe)
  real(rp)                        ,intent(in)       :: R(Nspe)
  real(rp)    ,allocatable        ,intent(out)      :: nu(:)
  real(rp)    ,allocatable        ,intent(out)      :: Ie(:)
  real(rp)    ,allocatable        ,intent(out)      :: Ia(:)
  real(rp)    ,allocatable        ,intent(out)      :: DnuL(:)
  real(rp)    ,allocatable        ,intent(out)      :: DnuD(:)
  integer                                           :: iSpe
  real(rp)                                          :: Nup
  real(rp)                                          :: Nlo
  real(rp)                                          :: Mav
  real(rp)                                          :: RMass
  real(rp)                                          :: Rad
  real(rp)    ,parameter                            :: IeParam  = hP*100.0_rp*c0/(4.0*pi)
  real(rp)    ,parameter                            :: IaParam  = 1.0_rp/(2.0E+8_rp*hP*c0**(2.0_rp))
  real(rp)    ,parameter                            :: DnuRes   = 1.2893E-19_rp
  real(rp)    ,parameter                            :: DnuVdW   = 1.95E-28_rp*(2.0_rp/c0)
  real(rp)    ,parameter                            :: DnuCol   = 2.0_rp/(c0*100.0_rp)
  real(rp)    ,allocatable                          :: Res(:)
  real(rp)    ,allocatable                          :: VdW(:)
  real(rp)    ,allocatable                          :: Col(:)
  
  Mav     = 0.0_rp
  do iSpe = 1, NSpe
    Mav   = M(iSpe)*X(iSpe) + Mav
  end do
  
  allocate(nu(NTotTra))
  allocate(Ie(NTotTra))
  allocate(Ia(NTotTra))
  allocate(DnuL(NTotTra))
  allocate(DnuD(NTotTra))
  
  nu      = LineParam(:,1)
  Ie      = IeParam*LineParam(:,1)*LineParam(:,2)*LineParam(:,3)*LineParam(:,4)*NoV*X(1)/Qtot
  Ia      = IaParam*(Ie(:)/(nu(:)**(3.0)))*( (LineParam(:,5)*LineParam(:,6)/(LineParam(:,3)*LineParam(:,4))) - 1.0_rp )
  
  allocate(Res(NTotTra))
  allocate(VdW(NTotTra))
  allocate(Col(NTotTra))
  
  Res     = DnuRes*sqrt(LineParam(:,3)/LineParam(:,5))*LineParam(:,2)*(LineParam(:,5)*LineParam(:,6)*NoV*X(1)/Qtot)/(nu(:)**(3.0))
  VdW     = DnuVdW*sqrt(2.0*Tg/Mav)*NoV*X(1)*(nu(:)**(2.0))
  Col     = 0.0_rp
  
  do iSpe = 1,NSpe
    RMass   = (M(1)/Na)*(M(iSpe)/Na)/( (M(1)/Na) + (M(iSpe)/Na) )
    Rad     = (R(1) + R(iSpe))*1E-10_rp
    Col(:)  = DnuCol*X(iSpe)*NoV*(Rad**(2.0))*sqrt( 8.0_rp*kB*pi*Tg/RMass ) + Col(:)
  end do
  
  DnuL    = Res + VdW + Col
  DnuD    = sqrt(8.0_rp*log(2.0_rp)*Ru*Tg/(c0**(2.0_rp)*M(1)))*nu

End Subroutine

Subroutine OutputData( NTotTra, nu, Ie, Ia, DnuL, DnuD, DecimParam )
  integer                         ,intent(in)       :: NTotTra
  real(rp)                        ,intent(in)       :: nu(NTotTra)
  real(rp)                        ,intent(in)       :: Ie(NTotTra)
  real(rp)                        ,intent(in)       :: Ia(NTotTra)
  real(rp)                        ,intent(in)       :: DnuL(NTotTra)
  real(rp)                        ,intent(in)       :: DnuD(NTotTra)
  real(rp)                        ,intent(in)       :: DecimParam
  character(*)  ,parameter                          :: OutFile = 'output.txt'
  integer                                           :: Unit
  integer                                           :: iTra
  integer                                           :: NTrueTra
  logical       ,allocatable                        :: DecimLog(:)
  logical       ,allocatable                        :: DecimLogIe(:)
  logical       ,allocatable                        :: DecimLogIa(:)
  real(rp)                                          :: MaxIe
  real(rp)                                          :: MaxIa
  real(rp)      ,allocatable                        :: realnu(:)
  real(rp)      ,allocatable                        :: realIe(:)
  real(rp)      ,allocatable                        :: realIa(:)
  real(rp)      ,allocatable                        :: realDnuL(:)
  real(rp)      ,allocatable                        :: realDnuD(:)
  
  allocate(DecimLogIe(NTotTra))
  allocate(DecimLogIa(NTotTra))
  
  MaxIe       = maxval(Ie(:))
  MaxIa       = maxval(Ia(:))
  
  DecimLogIe  = (Ie(:)/MaxIe >= DecimParam )
  DecimLogIa  = (Ia(:)/MaxIa >= DecimParam )
  
  DecimLog    = DecimLogIe .AND. DecimLogIa
  
  NTrueTra    = count(DecimLog)
  realnu      = pack(nu,DecimLog)
  realIe      = pack(Ie,DecimLog)
  realIa      = pack(Ia,DecimLog)
  realDnuL    = pack(DnuL,DecimLog)
  realDnuD    = pack(DnuD,DecimLog)
  
  write(*,'(a32,f9.5)') 'Percentage of decimated lines: ', (1.0 - real(NTrueTra)/real(NTotTra))*100.0
  write(*,'(a35,f9.5)') 'Percentage of original radiation: ', 100.0_rp*sum(realIe(:))/sum(Ie(:))
  
  open( Newunit=Unit, File=OutFile, ACTION='write', FORM='formatted' )
  write(Unit,'(i9)') NTrueTra
  do iTra = 1,NTrueTra
    write(Unit,'(5e16.8e3)') realnu(iTra), realIe(iTra), realIa(iTra), realDnuL(iTra), realDnuD(iTra)
  end do 
  close(Unit)
  
End Subroutine

Function ComputeLineParam( LinePos, VibTraCoe, Jup, l2up, Eup, Jlo, l2lo, Elo, Tr, Tv, NonEqInd, Nup, Nlo, Qtot ) result(LineParam)
  real(rp)                                          :: LinePos
  real(rp)                                          :: VibTraCoe(5)
  integer                                           :: Jup
  integer                                           :: l2up
  real(rp)                                          :: Eup(4)
  integer                                           :: Jlo
  integer                                           :: l2lo
  real(rp)                                          :: Elo(4)
  real(rp)                                          :: Tr
  real(rp)                                          :: Tv
  logical                                           :: NonEqInd
  real(rp)                                          :: Nup
  real(rp)                                          :: Nlo
  real(rp)                                          :: Qtot
  real(rp)                                          :: LineParam(6)
  integer                                           :: DeltaJ
  real(rp)                                          :: E
  real(rp)                                          :: HL
  real(rp)                                          :: HW
  real(rp)                                          :: Qr
  real(rp)                                          :: The
  real(rp)                                          :: The1
  real(rp)      ,parameter                          :: Er = 3.9021932090E-01_rp
  real(rp)      ,parameter                          :: c2_cmK = 100.0_rp*c2
  
  DeltaJ = Jup - Jlo
  select case (DeltaJ)
    case( 1 )
      HL = HonlLondonR( Jlo, l2up-l2lo, l2lo )
      HW = HermanWallisR( Jlo, VibTraCoe )
    case( 0 )
      HL = HonlLondonQ( Jlo, l2up-l2lo, l2lo )
      HW = HermanWallisQ( Jlo, VibTraCoe )
    case( -1 )
      HL = HonlLondonP( Jlo, l2up-l2lo, l2lo )
      HW = HermanWallisP( Jlo, VibTraCoe )
  end select
  LineParam(1)  = LinePos
  LineParam(2)  = VibTraCoe(1)*HL*HW
  LineParam(3)  = 2.0*real(Jup) + 1.0
  E             = Energy( Eup, Jup)
  LineParam(4)  = exp( -(E-Eup(1))*100.0_rp*c2/Tr )*exp( -Eup(1)*100.0_rp*c2/Tv )
  LineParam(5)  = 2.0*real(Jlo) + 1.0
  E             = Energy( Elo, Jlo)
  LineParam(6)  = exp( -(E-Elo(1))*100.0_rp*c2/Tr )*exp( -Elo(1)*100.0_rp*c2/Tv )
  
  if( NonEqInd ) then
    The           = Tr/(c2_cmK*Er)
    The1          = 1.0/The
    Qr            = (The/2.0_rp)*exp( The1/3.0_rp + The1*The1/90.0 )
    
    E             = Energy( Eup, Jup)
    LineParam(3)  = 2.0*real(Jup) + 1.0
    LineParam(4)  = Nup*(exp( -(E-Eup(1))*100.0_rp*c2/Tr )/Qr)*(Qtot/(LineParam(3)*NoV)) !
    
    E             = Energy( Elo, Jlo)
    LineParam(5)  = 2.0*real(Jlo) + 1.0
    LineParam(6)  = Nlo*(exp( -(E-Elo(1))*100.0_rp*c2/Tr )/Qr)*(Qtot/(LineParam(5)*NoV))
  end if
  
End Function

Function Energy( Ecoe, J ) result(E)
  real(rp)                                          :: Ecoe(4)
  integer                                           :: J
  real(rp)                                          :: E
  real(rp)                                          :: realJ
  real(rp)                                          :: X
  realJ = real(J)
  X     = realJ*(realJ+1.0_rp)
  E     = Ecoe(1) + Ecoe(2)*X - Ecoe(3)*X*X + Ecoe(4)*X*X*X
End Function

Function ChangeJ( J, l2, step ) result(newJ)
  integer                                           :: J
  integer                                           :: l2
  integer                                           :: step
  integer                                           :: newJ
  integer                                           :: dJ
  if( l2 == 0 ) then
    dJ = 2
  else 
    dJ = 1
  end if
  newJ = J + step*dJ
End Function

Function HonlLondonP( Jpp, Deltal, lpp ) result(HL)
  integer                              ,intent(in)  ::  Jpp
  integer                              ,intent(in)  ::  Deltal
  integer                              ,intent(in)  ::  lpp
  real(rp)                                          ::  HL
  real(rp)                                          ::  rJpp
  real(rp)                                          ::  rDeltal
  real(rp)                                          ::  rlpp
  real(rp)                                          ::  HL1, HL2
  rJpp    = real(Jpp)
  rlpp    = real(lpp)
  rDeltal = real(Deltal)
  select case ( Deltal )
    case( 0 )
      HL1 = ( rJpp + rlpp )*( rJpp - rlpp )
      HL2 = ( rJpp )*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
    case default
      HL1 = ( rJpp - 1.0_rp - rlpp*rDeltal )*( rJpp - rlpp*rDeltal )
      HL2 = ( 2.0_rp*rJpp )*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
  end select
  if( HL2 == 0.0_rp ) HL = 0.0_rp
End Function

Function HonlLondonQ( Jpp, Deltal, lpp ) result(HL)
  integer                              ,intent(in)  ::  Jpp
  integer                              ,intent(in)  ::  Deltal
  integer                              ,intent(in)  ::  lpp
  real(rp)                                          ::  HL
  real(rp)                                          ::  rJpp
  real(rp)                                          ::  rDeltal
  real(rp)                                          ::  rlpp
  real(rp)                                          ::  HL1, HL2
  rJpp    = real(Jpp)
  rlpp    = real(lpp)
  rDeltal = real(Deltal)
  select case ( Deltal )
    case( 0 )
      HL1 = ( 2.0_rp*rJpp + 1.0_rp )*(rlpp**2.0_rp)
      HL2 = ( rJpp*( rJpp + 1.0_rp ) )*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
    case default
      HL1 = ( rJpp + 1.0_rp + rlpp*rDeltal )*( rJpp - rlpp*rDeltal )*( 2.0_rp*rJpp + 1.0_rp )
      HL2 = ( 2.0_rp*rJpp*( rJpp + 1.0_rp ) )*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
  end select
  if( HL2 == 0.0_rp ) HL = 0.0_rp
End Function

Function HonlLondonR( Jpp, Deltal, lpp ) result(HL)
  integer                              ,intent(in)  ::  Jpp
  integer                              ,intent(in)  ::  Deltal
  integer                              ,intent(in)  ::  lpp
  real(rp)                                          ::  HL
  real(rp)                                          ::  rJpp
  real(rp)                                          ::  rDeltal
  real(rp)                                          ::  rlpp
  real(rp)                                          ::  HL1, HL2
  rJpp    = real(Jpp)
  rlpp    = real(lpp)
  rDeltal = real(Deltal)
  select case ( Deltal )
    case( 0 )
      HL1 = ( rJpp + 1.0_rp + rlpp )*( rJpp + 1.0_rp - rlpp )
      HL2 = ( rJpp + 1.0_rp)*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
    case default
      HL1 = ( rJpp + 2.0_rp + rlpp*rDeltal )*( rJpp + 1.0_rp + rlpp*rDeltal )
      HL2 = (2.0_rp*( rJpp + 1.0_rp ))*(2.0_rp*rJpp + 1.0_rp)
      HL  = HL1/HL2
  end select
  if( HL2 == 0.0_rp ) HL = 0.0_rp
End Function

Function HermanWallisP( Jlo, HWCoe ) result(HW)
  integer                                           :: Jlo
  real(rp)                                          :: HWCoe(5)
  real(rp)                                          :: X
  real(rp)                                          :: HW
  
  X   = real(Jlo)
  HW  = ( 1.0_rp - HWCoe(2)*X + HWCoe(3)*X*X - HWCoe(4)*X*X*X )**2.0_rp

End Function

Function HermanWallisQ( Jlo, HWCoe ) result(HW)
  integer                                           :: Jlo
  real(rp)                                          :: HWCoe(5)
  real(rp)                                          :: X
  real(rp)                                          :: HW
  
  X   = real(Jlo)*(real(Jlo) + 1.0_rp)
  HW  = ( 1.0_rp + HWCoe(5)*X )**2.0_rp

End Function

Function HermanWallisR( Jlo, HWCoe ) result(HW)
  integer                                           :: Jlo
  real(rp)                                          :: HWCoe(5)
  real(rp)                                          :: X
  real(rp)                                          :: HW
  
  X   = (real(Jlo) + 1.0_rp)
  HW  = ( 1.0_rp + HWCoe(2)*X + HWCoe(3)*X*X + HWCoe(4)*X*X*X )**2.0_rp

End Function
  
End Program
